@@ -1,8 +1,8 @@
 /*
  *
- * Copyright (c) 2014-2021 The Khronos Group Inc.
- * Copyright (c) 2014-2021 Valve Corporation
- * Copyright (c) 2014-2021 LunarG, Inc.
+ * Copyright (c) 2014-2022 The Khronos Group Inc.
+ * Copyright (c) 2014-2022 Valve Corporation
+ * Copyright (c) 2014-2022 LunarG, Inc.
  * Copyright (C) 2015 Google Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -168,7 +168,7 @@ bool windows_get_device_registry_entry(const struct loader_instance *inst, char
 
     CONFIGRET status = CM_Open_DevNode_Key(dev_id, KEY_QUERY_VALUE, 0, RegDisposition_OpenExisting, &hkrKey, CM_REGISTRY_SOFTWARE);
     if (status != CR_SUCCESS) {
-        loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
+        loader_log(inst, VULKAN_LOADER_WARN_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
                    "windows_get_device_registry_entry: Failed to open registry key for DeviceID(%d)", dev_id);
         *result = VK_ERROR_INCOMPATIBLE_DRIVER;
         return false;
@@ -179,18 +179,19 @@ bool windows_get_device_registry_entry(const struct loader_instance *inst, char
 
     if (ret != ERROR_SUCCESS) {
         if (ret == ERROR_FILE_NOT_FOUND) {
-            loader_log(inst, VULKAN_LOADER_INFO_BIT, 0,
+            loader_log(inst, VULKAN_LOADER_INFO_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
                        "windows_get_device_registry_entry: Device ID(%d) Does not contain a value for \"%s\"", dev_id, value_name);
         } else {
-            loader_log(inst, VULKAN_LOADER_INFO_BIT, 0, "windows_get_device_registry_entry: DeviceID(%d) Failed to obtain %s size",
-                       dev_id, value_name);
+            loader_log(inst, VULKAN_LOADER_INFO_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
+                       "windows_get_device_registry_entry: DeviceID(%d) Failed to obtain %s size", dev_id, value_name);
         }
         goto out;
     }
 
     manifest_path = loader_instance_heap_alloc(inst, requiredSize, VK_SYSTEM_ALLOCATION_SCOPE_INSTANCE);
     if (manifest_path == NULL) {
-        loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0, "windows_get_device_registry_entry: Failed to allocate space for DriverName.");
+        loader_log(inst, VULKAN_LOADER_ERROR_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
+                   "windows_get_device_registry_entry: Failed to allocate space for DriverName.");
         *result = VK_ERROR_OUT_OF_HOST_MEMORY;
         goto out;
     }
@@ -198,14 +199,14 @@ bool windows_get_device_registry_entry(const struct loader_instance *inst, char
     ret = RegQueryValueEx(hkrKey, value_name, NULL, &data_type, (BYTE *)manifest_path, &requiredSize);
 
     if (ret != ERROR_SUCCESS) {
-        loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0, "windows_get_device_registry_entry: DeviceID(%d) Failed to obtain %s",
-                   value_name);
+        loader_log(inst, VULKAN_LOADER_ERROR_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
+                   "windows_get_device_registry_entry: DeviceID(%d) Failed to obtain %s", value_name);
         *result = VK_ERROR_INCOMPATIBLE_DRIVER;
         goto out;
     }
 
     if (data_type != REG_SZ && data_type != REG_MULTI_SZ) {
-        loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0,
+        loader_log(inst, VULKAN_LOADER_ERROR_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
                    "windows_get_device_registry_entry: Invalid %s data type. Expected REG_SZ or REG_MULTI_SZ.", value_name);
         *result = VK_ERROR_INCOMPATIBLE_DRIVER;
         goto out;
@@ -221,8 +222,8 @@ out:
     return found;
 }
 
-VkResult windows_get_device_registry_files(const struct loader_instance *inst, char **reg_data, PDWORD reg_data_size,
-                                           LPCSTR value_name) {
+VkResult windows_get_device_registry_files(const struct loader_instance *inst, uint32_t log_target_flag, char **reg_data,
+                                           PDWORD reg_data_size, LPCSTR value_name) {
     static const wchar_t *softwareComponentGUID = L"{5c4c3332-344d-483c-8739-259e934c9cc8}";
     static const wchar_t *displayGUID = L"{4d36e968-e325-11ce-bfc1-08002be10318}";
 #ifdef CM_GETIDLIST_FILTER_PRESENT
@@ -232,6 +233,9 @@ VkResult windows_get_device_registry_files(const struct loader_instance *inst, c
 #endif
 
     wchar_t childGuid[MAX_GUID_STRING_LEN + 2];  // +2 for brackets {}
+    for (uint32_t i = 0; i < MAX_GUID_STRING_LEN + 2; i++) {
+        childGuid[i] = L'\0';
+    }
     ULONG childGuidSize = sizeof(childGuid);
 
     DEVINST devID = 0, childID = 0;
@@ -252,7 +256,7 @@ VkResult windows_get_device_registry_files(const struct loader_instance *inst, c
 
         pDeviceNames = loader_instance_heap_alloc(inst, deviceNamesSize * sizeof(wchar_t), VK_SYSTEM_ALLOCATION_SCOPE_INSTANCE);
         if (pDeviceNames == NULL) {
-            loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0,
+            loader_log(inst, VULKAN_LOADER_ERROR_BIT | log_target_flag, 0,
                        "windows_get_device_registry_files: Failed to allocate space for display device names.");
             result = VK_ERROR_OUT_OF_HOST_MEMORY;
             return result;
@@ -263,25 +267,26 @@ VkResult windows_get_device_registry_files(const struct loader_instance *inst, c
         for (wchar_t *deviceName = pDeviceNames; *deviceName; deviceName += wcslen(deviceName) + 1) {
             CONFIGRET status = CM_Locate_DevNodeW(&devID, deviceName, CM_LOCATE_DEVNODE_NORMAL);
             if (CR_SUCCESS != status) {
-                loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0, "windows_get_device_registry_files: failed to open DevNode %ls",
-                           deviceName);
+                loader_log(inst, VULKAN_LOADER_ERROR_BIT | log_target_flag, 0,
+                           "windows_get_device_registry_files: failed to open DevNode %ls", deviceName);
                 continue;
             }
             ULONG ulStatus, ulProblem;
             status = CM_Get_DevNode_Status(&ulStatus, &ulProblem, devID, 0);
 
             if (CR_SUCCESS != status) {
-                loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0, "windows_get_device_registry_files: failed to probe device status %ls",
-                           deviceName);
+                loader_log(inst, VULKAN_LOADER_ERROR_BIT | log_target_flag, 0,
+                           "windows_get_device_registry_files: failed to probe device status %ls", deviceName);
                 continue;
             }
             if ((ulStatus & DN_HAS_PROBLEM) && (ulProblem == CM_PROB_NEED_RESTART || ulProblem == DN_NEED_RESTART)) {
-                loader_log(inst, VULKAN_LOADER_INFO_BIT, 0,
+                loader_log(inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0,
                            "windows_get_device_registry_files: device %ls is pending reboot, skipping ...", deviceName);
                 continue;
             }
 
-            loader_log(inst, VULKAN_LOADER_INFO_BIT, 0, "windows_get_device_registry_files: opening device %ls", deviceName);
+            loader_log(inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0, "windows_get_device_registry_files: opening device %ls",
+                       deviceName);
 
             if (windows_get_device_registry_entry(inst, reg_data, reg_data_size, devID, value_name, &result)) {
                 found = true;
@@ -292,7 +297,7 @@ VkResult windows_get_device_registry_files(const struct loader_instance *inst, c
 
             status = CM_Get_Child(&childID, devID, 0);
             if (status != CR_SUCCESS) {
-                loader_log(inst, VULKAN_LOADER_INFO_BIT, 0,
+                loader_log(inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0,
                            "windows_get_device_registry_files: unable to open child-device error:%d", status);
                 continue;
             }
@@ -301,12 +306,12 @@ VkResult windows_get_device_registry_files(const struct loader_instance *inst, c
                 wchar_t buffer[MAX_DEVICE_ID_LEN];
                 CM_Get_Device_IDW(childID, buffer, MAX_DEVICE_ID_LEN, 0);
 
-                loader_log(inst, VULKAN_LOADER_INFO_BIT, 0, "windows_get_device_registry_files: Opening child device %d - %ls",
-                           childID, buffer);
+                loader_log(inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0,
+                           "windows_get_device_registry_files: Opening child device %d - %ls", childID, buffer);
 
                 status = CM_Get_DevNode_Registry_PropertyW(childID, CM_DRP_CLASSGUID, NULL, &childGuid, &childGuidSize, 0);
                 if (status != CR_SUCCESS) {
-                    loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0,
+                    loader_log(inst, VULKAN_LOADER_ERROR_BIT | log_target_flag, 0,
                                "windows_get_device_registry_files: unable to obtain GUID for:%d error:%d", childID, status);
 
                     result = VK_ERROR_INCOMPATIBLE_DRIVER;
@@ -314,7 +319,7 @@ VkResult windows_get_device_registry_files(const struct loader_instance *inst, c
                 }
 
                 if (wcscmp(childGuid, softwareComponentGUID) != 0) {
-                    loader_log(inst, VULKAN_LOADER_DEBUG_BIT, 0,
+                    loader_log(inst, VULKAN_LOADER_DEBUG_BIT | log_target_flag, 0,
                                "windows_get_device_registry_files: GUID for %d is not SoftwareComponent skipping", childID);
                     continue;
                 }
@@ -331,6 +336,7 @@ VkResult windows_get_device_registry_files(const struct loader_instance *inst, c
     }
 
     if (!found && result != VK_ERROR_OUT_OF_HOST_MEMORY) {
+        loader_log(inst, log_target_flag, 0, "windows_get_device_registry_files: found no registry files");
         result = VK_ERROR_INCOMPATIBLE_DRIVER;
     }
 
@@ -395,13 +401,14 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
     bool found = false;
     IDXGIFactory1 *dxgi_factory = NULL;
     bool is_driver = !strcmp(location, VK_DRIVERS_INFO_REGISTRY_LOC);
+    uint32_t log_target_flag = is_driver ? VULKAN_LOADER_DRIVER_BIT : VULKAN_LOADER_LAYER_BIT;
 
     assert(reg_data != NULL && "windows_get_registry_files: reg_data is a NULL pointer");
 
     if (is_driver) {
         HRESULT hres = fpCreateDXGIFactory1(&IID_IDXGIFactory1, (void **)&dxgi_factory);
         if (hres != S_OK) {
-            loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
+            loader_log(inst, VULKAN_LOADER_WARN_BIT | log_target_flag, 0,
                        "windows_get_registry_files: Failed to create dxgi factory for ICD registry verification. No ICDs will be "
                        "added from "
                        "legacy registry locations");
@@ -421,7 +428,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
                     if (NULL == *reg_data) {
                         *reg_data = loader_instance_heap_alloc(inst, *reg_data_size, VK_SYSTEM_ALLOCATION_SCOPE_INSTANCE);
                         if (NULL == *reg_data) {
-                            loader_log(inst, VULKAN_LOADER_ERROR_BIT, 0,
+                            loader_log(inst, VULKAN_LOADER_ERROR_BIT | log_target_flag, 0,
                                        "windows_get_registry_files: Failed to allocate space for registry data for key %s", name);
                             RegCloseKey(key);
                             result = VK_ERROR_OUT_OF_HOST_MEMORY;
@@ -433,7 +440,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
                                                                      VK_SYSTEM_ALLOCATION_SCOPE_INSTANCE);
                         if (NULL == new_ptr) {
                             loader_log(
-                                inst, VULKAN_LOADER_ERROR_BIT, 0,
+                                inst, VULKAN_LOADER_ERROR_BIT | log_target_flag, 0,
                                 "windows_get_registry_files: Failed to reallocate space for registry value of size %d for key %s",
                                 *reg_data_size * 2, name);
                             RegCloseKey(key);
@@ -446,7 +453,8 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
 
                     // We've now found a json file. If this is an ICD, we still need to check if there is actually a device
                     // that matches this ICD
-                    loader_log(inst, VULKAN_LOADER_INFO_BIT, 0, "Located json file \"%s\" from registry \"%s\\%s\"", name,
+                    loader_log(inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0,
+                               "Located json file \"%s\" from registry \"%s\\%s\"", name,
                                hive == DEFAULT_VK_REGISTRY_HIVE ? DEFAULT_VK_REGISTRY_HIVE_STR : SECONDARY_VK_REGISTRY_HIVE_STR,
                                location);
                     if (is_driver) {
@@ -457,7 +465,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
                             }
                         }
                         if (i == sizeof(known_drivers) / sizeof(known_drivers[0])) {
-                            loader_log(inst, VULKAN_LOADER_INFO_BIT, 0,
+                            loader_log(inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0,
                                        "Driver %s is not recognized as a known driver. It will be assumed to be active", name);
                         } else {
                             bool found_gpu = false;
@@ -467,7 +475,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
                                 if (hres == DXGI_ERROR_NOT_FOUND) {
                                     break;
                                 } else if (hres != S_OK) {
-                                    loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
+                                    loader_log(inst, VULKAN_LOADER_WARN_BIT | log_target_flag, 0,
                                                "Failed to enumerate DXGI adapters at index %d. As a result, drivers may be skipped",
                                                j);
                                     continue;
@@ -477,7 +485,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
                                 hres = adapter->lpVtbl->GetDesc1(adapter, &description);
                                 if (hres != S_OK) {
                                     loader_log(
-                                        inst, VULKAN_LOADER_INFO_BIT, 0,
+                                        inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0,
                                         "Failed to get DXGI adapter information at index %d. As a result, drivers may be skipped",
                                         j);
                                     continue;
@@ -490,7 +498,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
                             }
 
                             if (!found_gpu) {
-                                loader_log(inst, VULKAN_LOADER_INFO_BIT, 0,
+                                loader_log(inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0,
                                            "Dropping driver %s as no corresponding DXGI adapter was found", name);
                                 continue;
                             }
@@ -523,7 +531,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
                             found = true;
                         } else {
                             loader_log(
-                                inst, VULKAN_LOADER_INFO_BIT, 0,
+                                inst, VULKAN_LOADER_INFO_BIT | log_target_flag, 0,
                                 "Skipping adding of json file \"%s\" from registry \"%s\\%s\" to the list due to duplication", name,
                                 hive == DEFAULT_VK_REGISTRY_HIVE ? DEFAULT_VK_REGISTRY_HIVE_STR : SECONDARY_VK_REGISTRY_HIVE_STR,
                                 location);
@@ -544,6 +552,7 @@ VkResult windows_get_registry_files(const struct loader_instance *inst, char *lo
     }
 
     if (!found && result != VK_ERROR_OUT_OF_HOST_MEMORY) {
+        loader_log(inst, log_target_flag, 0, "Found no registry files in %s", location);
         result = VK_ERROR_INCOMPATIBLE_DRIVER;
     }
 
@@ -565,10 +574,15 @@ VkResult windows_read_manifest_from_d3d_adapters(const struct loader_instance *i
     char *json_path = NULL;
     size_t json_path_size = 0;
 
-    PFN_LoaderEnumAdapters2 fpLoaderEnumAdapters2 =
-        (PFN_LoaderEnumAdapters2)GetProcAddress(GetModuleHandle("gdi32.dll"), "D3DKMTEnumAdapters2");
+    HMODULE gdi32_dll = GetModuleHandle("gdi32.dll");
+    if (gdi32_dll == NULL) {
+        result = VK_ERROR_INCOMPATIBLE_DRIVER;
+        goto out;
+    }
+
+    PFN_LoaderEnumAdapters2 fpLoaderEnumAdapters2 = (PFN_LoaderEnumAdapters2)GetProcAddress(gdi32_dll, "D3DKMTEnumAdapters2");
     PFN_LoaderQueryAdapterInfo fpLoaderQueryAdapterInfo =
-        (PFN_LoaderQueryAdapterInfo)GetProcAddress(GetModuleHandle("gdi32.dll"), "D3DKMTQueryAdapterInfo");
+        (PFN_LoaderQueryAdapterInfo)GetProcAddress(gdi32_dll, "D3DKMTQueryAdapterInfo");
     if (fpLoaderEnumAdapters2 == NULL || fpLoaderQueryAdapterInfo == NULL) {
         result = VK_ERROR_INCOMPATIBLE_DRIVER;
         goto out;
@@ -693,13 +707,14 @@ VkResult windows_read_data_files_in_registry(const struct loader_instance *inst,
                                              struct loader_data_files *out_files) {
     VkResult vk_result = VK_SUCCESS;
     char *search_path = NULL;
+    uint32_t log_target_flag = 0;
 
-    if (data_file_type == LOADER_DATA_FILE_MANIFEST_ICD) {
-        loader_log(inst, VULKAN_LOADER_DRIVER_BIT, 0, "Checking for Driver Manifest files in Registry at %s",
-                   registry_location);
+    if (data_file_type == LOADER_DATA_FILE_MANIFEST_DRIVER) {
+        log_target_flag = VULKAN_LOADER_DRIVER_BIT;
+        loader_log(inst, log_target_flag, 0, "Checking for Driver Manifest files in Registry at %s", registry_location);
     } else {
-        loader_log(inst, VULKAN_LOADER_LAYER_BIT, 0, "Checking for Driver Manifest files in Registry at %s",
-                   registry_location);
+        log_target_flag = VULKAN_LOADER_LAYER_BIT;
+        loader_log(inst, log_target_flag, 0, "Checking for Layer Manifest files in Registry at %s", registry_location);
     }
 
     // These calls look at the PNP/Device section of the registry.
@@ -709,17 +724,20 @@ VkResult windows_read_data_files_in_registry(const struct loader_instance *inst,
         // If we're looking for drivers we need to try enumerating adapters
         regHKR_result = windows_read_manifest_from_d3d_adapters(inst, &search_path, &reg_size, LoaderPnpDriverRegistryWide());
         if (regHKR_result == VK_INCOMPLETE) {
-            regHKR_result = windows_get_device_registry_files(inst, &search_path, &reg_size, LoaderPnpDriverRegistry());
+            regHKR_result =
+                windows_get_device_registry_files(inst, log_target_flag, &search_path, &reg_size, LoaderPnpDriverRegistry());
         }
     } else if (!strncmp(registry_location, VK_ELAYERS_INFO_REGISTRY_LOC, sizeof(VK_ELAYERS_INFO_REGISTRY_LOC))) {
         regHKR_result = windows_read_manifest_from_d3d_adapters(inst, &search_path, &reg_size, LoaderPnpELayerRegistryWide());
         if (regHKR_result == VK_INCOMPLETE) {
-            regHKR_result = windows_get_device_registry_files(inst, &search_path, &reg_size, LoaderPnpELayerRegistry());
+            regHKR_result =
+                windows_get_device_registry_files(inst, log_target_flag, &search_path, &reg_size, LoaderPnpELayerRegistry());
         }
     } else if (!strncmp(registry_location, VK_ILAYERS_INFO_REGISTRY_LOC, sizeof(VK_ILAYERS_INFO_REGISTRY_LOC))) {
         regHKR_result = windows_read_manifest_from_d3d_adapters(inst, &search_path, &reg_size, LoaderPnpILayerRegistryWide());
         if (regHKR_result == VK_INCOMPLETE) {
-            regHKR_result = windows_get_device_registry_files(inst, &search_path, &reg_size, LoaderPnpILayerRegistry());
+            regHKR_result =
+                windows_get_device_registry_files(inst, log_target_flag, &search_path, &reg_size, LoaderPnpILayerRegistry());
         }
     }
 
@@ -728,8 +746,8 @@ VkResult windows_read_data_files_in_registry(const struct loader_instance *inst,
         goto out;
     }
 
-    // This call looks into the Khronos non-device specific section of the registry.
-    bool use_secondary_hive = (data_file_type == LOADER_DATA_FILE_MANIFEST_LAYER) && (!is_high_integrity());
+    // This call looks into the Khronos non-device specific section of the registry for layer files.
+    bool use_secondary_hive = (data_file_type != LOADER_DATA_FILE_MANIFEST_DRIVER) && (!is_high_integrity());
     VkResult reg_result = windows_get_registry_files(inst, registry_location, use_secondary_hive, &search_path, &reg_size);
     if (reg_result == VK_ERROR_OUT_OF_HOST_MEMORY) {
         vk_result = VK_ERROR_OUT_OF_HOST_MEMORY;
@@ -737,20 +755,21 @@ VkResult windows_read_data_files_in_registry(const struct loader_instance *inst,
     }
 
     if ((VK_SUCCESS != reg_result && VK_SUCCESS != regHKR_result) || NULL == search_path) {
-        if (data_file_type == LOADER_DATA_FILE_MANIFEST_ICD) {
-            loader_log(inst, VULKAN_LOADER_ERROR_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
+        if (data_file_type == LOADER_DATA_FILE_MANIFEST_DRIVER) {
+            loader_log(inst, VULKAN_LOADER_ERROR_BIT | log_target_flag, 0,
                        "windows_read_data_files_in_registry: Registry lookup failed to get ICD manifest files.  Possibly missing "
                        "Vulkan driver?");
             vk_result = VK_ERROR_INCOMPATIBLE_DRIVER;
         } else {
             if (warn_if_not_present) {
-                if (data_file_type == LOADER_DATA_FILE_MANIFEST_LAYER) {
+                if (data_file_type == LOADER_DATA_FILE_MANIFEST_IMPLICIT_LAYER ||
+                    data_file_type == LOADER_DATA_FILE_MANIFEST_EXPLICIT_LAYER) {
                     // This is only a warning for layers
-                    loader_log(inst, VULKAN_LOADER_WARN_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
+                    loader_log(inst, VULKAN_LOADER_WARN_BIT | log_target_flag, 0,
                                "windows_read_data_files_in_registry: Registry lookup failed to get layer manifest files.");
                 } else {
                     // This is only a warning for general data files
-                    loader_log(inst, VULKAN_LOADER_WARN_BIT | VULKAN_LOADER_DRIVER_BIT, 0,
+                    loader_log(inst, VULKAN_LOADER_WARN_BIT | log_target_flag, 0,
                                "windows_read_data_files_in_registry: Registry lookup failed to get data files.");
                 }
             }
@@ -761,7 +780,7 @@ VkResult windows_read_data_files_in_registry(const struct loader_instance *inst,
     }
 
     // Now, parse the paths and add any manifest files found in them.
-    vk_result = add_data_files_in_path(inst, search_path, false, out_files, false);
+    vk_result = add_data_files(inst, search_path, out_files, false);
 
 out:
 
@@ -773,8 +792,8 @@ out:
 }
 
 // This function allocates an array in sorted_devices which must be freed by the caller if not null
-VkResult windows_read_sorted_physical_devices(struct loader_instance *inst, struct LoaderSortedPhysicalDevice **sorted_devices,
-                                              uint32_t *sorted_count) {
+VkResult windows_read_sorted_physical_devices(struct loader_instance *inst, uint32_t *sorted_devices_count,
+                                              struct loader_phys_dev_per_icd **sorted_devices) {
     VkResult res = VK_SUCCESS;
 
     uint32_t sorted_alloc = 0;
@@ -783,121 +802,119 @@ VkResult windows_read_sorted_physical_devices(struct loader_instance *inst, stru
     HRESULT hres = fpCreateDXGIFactory1(&IID_IDXGIFactory6, (void **)&dxgi_factory);
     if (hres != S_OK) {
         loader_log(inst, VULKAN_LOADER_INFO_BIT, 0, "Failed to create DXGI factory 6. Physical devices will not be sorted");
-    } else {
-        sorted_alloc = 16;
-        *sorted_devices = loader_instance_heap_alloc(inst, sorted_alloc * sizeof(struct LoaderSortedPhysicalDevice),
-                                                     VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
-        if (*sorted_devices == NULL) {
-            res = VK_ERROR_OUT_OF_HOST_MEMORY;
-            goto out;
+        goto out;
+    }
+    sorted_alloc = 16;
+    *sorted_devices =
+        loader_instance_heap_alloc(inst, sorted_alloc * sizeof(struct loader_phys_dev_per_icd), VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
+    if (*sorted_devices == NULL) {
+        res = VK_ERROR_OUT_OF_HOST_MEMORY;
+        goto out;
+    }
+
+    memset(*sorted_devices, 0, sorted_alloc * sizeof(struct loader_phys_dev_per_icd));
+
+    for (uint32_t i = 0;; ++i) {
+        IDXGIAdapter1 *adapter;
+        hres = dxgi_factory->lpVtbl->EnumAdapterByGpuPreference(dxgi_factory, i, DXGI_GPU_PREFERENCE_UNSPECIFIED,
+                                                                &IID_IDXGIAdapter1, (void **)&adapter);
+        if (hres == DXGI_ERROR_NOT_FOUND) {
+            break;  // No more adapters
+        } else if (hres != S_OK) {
+            loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
+                       "Failed to enumerate adapters by GPU preference at index %u. This adapter will not be sorted", i);
+            break;
         }
 
-        memset(*sorted_devices, 0, sorted_alloc * sizeof(struct LoaderSortedPhysicalDevice));
+        DXGI_ADAPTER_DESC1 description;
+        hres = adapter->lpVtbl->GetDesc1(adapter, &description);
+        if (hres != S_OK) {
+            loader_log(inst, VULKAN_LOADER_WARN_BIT, 0, "Failed to get adapter LUID index %u. This adapter will not be sorted", i);
+            continue;
+        }
 
-        *sorted_count = 0;
-        for (uint32_t i = 0;; ++i) {
-            IDXGIAdapter1 *adapter;
-            hres = dxgi_factory->lpVtbl->EnumAdapterByGpuPreference(dxgi_factory, i, DXGI_GPU_PREFERENCE_UNSPECIFIED,
-                                                                    &IID_IDXGIAdapter1, (void **)&adapter);
-            if (hres == DXGI_ERROR_NOT_FOUND) {
-                break;  // No more adapters
-            } else if (hres != S_OK) {
-                loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
-                           "Failed to enumerate adapters by GPU preference at index %u. This adapter will not be sorted", i);
-                break;
+        if (sorted_alloc <= i) {
+            uint32_t old_size = sorted_alloc * sizeof(struct loader_phys_dev_per_icd);
+            *sorted_devices =
+                loader_instance_heap_realloc(inst, *sorted_devices, old_size, 2 * old_size, VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
+            if (*sorted_devices == NULL) {
+                adapter->lpVtbl->Release(adapter);
+                res = VK_ERROR_OUT_OF_HOST_MEMORY;
+                goto out;
             }
+            sorted_alloc *= 2;
+        }
+        struct loader_phys_dev_per_icd *sorted_array = *sorted_devices;
+        sorted_array[*sorted_devices_count].device_count = 0;
+        sorted_array[*sorted_devices_count].physical_devices = NULL;
 
-            DXGI_ADAPTER_DESC1 description;
-            hres = adapter->lpVtbl->GetDesc1(adapter, &description);
-            if (hres != S_OK) {
-                loader_log(inst, VULKAN_LOADER_WARN_BIT, 0, "Failed to get adapter LUID index %u. This adapter will not be sorted",
-                           i);
+        icd_term = inst->icd_terms;
+        for (uint32_t icd_idx = 0; NULL != icd_term; icd_term = icd_term->next, icd_idx++) {
+            // This is the new behavior, which cannot be run unless the ICD provides EnumerateAdapterPhysicalDevices
+            if (icd_term->scanned_icd->EnumerateAdapterPhysicalDevices == NULL) {
                 continue;
             }
 
-            if (sorted_alloc <= i) {
-                uint32_t old_size = sorted_alloc * sizeof(struct LoaderSortedPhysicalDevice);
-                *sorted_devices =
-                    loader_instance_heap_realloc(inst, *sorted_devices, old_size, 2 * old_size, VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
-                if (*sorted_devices == NULL) {
-                    adapter->lpVtbl->Release(adapter);
-                    res = VK_ERROR_OUT_OF_HOST_MEMORY;
-                    goto out;
-                }
-                sorted_alloc *= 2;
+            uint32_t count = 0;
+            VkResult vkres =
+                icd_term->scanned_icd->EnumerateAdapterPhysicalDevices(icd_term->instance, description.AdapterLuid, &count, NULL);
+            if (vkres == VK_ERROR_INCOMPATIBLE_DRIVER) {
+                continue;  // This driver doesn't support the adapter
+            } else if (vkres == VK_ERROR_OUT_OF_HOST_MEMORY) {
+                res = VK_ERROR_OUT_OF_HOST_MEMORY;
+                goto out;
+            } else if (vkres != VK_SUCCESS) {
+                loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
+                           "Failed to convert DXGI adapter into Vulkan physical device with unexpected error code");
+                continue;
             }
-            struct LoaderSortedPhysicalDevice *sorted_array = *sorted_devices;
-            sorted_array[*sorted_count].device_count = 0;
-            sorted_array[*sorted_count].physical_devices = NULL;
-            //*sorted_count = i;
 
-            icd_term = inst->icd_terms;
-            for (uint32_t icd_idx = 0; NULL != icd_term; icd_term = icd_term->next, icd_idx++) {
-                // This is the new behavior, which cannot be run unless the ICD provides EnumerateAdapterPhysicalDevices
-                if (icd_term->scanned_icd->EnumerateAdapterPhysicalDevices == NULL) {
-                    continue;
-                }
-
-                uint32_t count;
-                VkResult vkres = icd_term->scanned_icd->EnumerateAdapterPhysicalDevices(icd_term->instance, description.AdapterLuid,
-                                                                                        &count, NULL);
-                if (vkres == VK_ERROR_INCOMPATIBLE_DRIVER) {
-                    continue;  // This driver doesn't support the adapter
-                } else if (vkres == VK_ERROR_OUT_OF_HOST_MEMORY) {
-                    res = VK_ERROR_OUT_OF_HOST_MEMORY;
-                    goto out;
-                } else if (vkres != VK_SUCCESS) {
-                    loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
-                               "Failed to convert DXGI adapter into Vulkan physical device with unexpected error code");
-                    continue;
-                }
-
-                // Get the actual physical devices
-                if (0 != count) {
-                    do {
-                        sorted_array[*sorted_count].physical_devices =
-                            loader_instance_heap_realloc(inst, sorted_array[*sorted_count].physical_devices,
-                                                         sorted_array[*sorted_count].device_count * sizeof(VkPhysicalDevice),
-                                                         count * sizeof(VkPhysicalDevice), VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
-                        if (sorted_array[*sorted_count].physical_devices == NULL) {
-                            res = VK_ERROR_OUT_OF_HOST_MEMORY;
-                            break;
-                        }
-                        sorted_array[*sorted_count].device_count = count;
-                    } while ((vkres = icd_term->scanned_icd->EnumerateAdapterPhysicalDevices(
-                                  icd_term->instance, description.AdapterLuid, &count,
-                                  sorted_array[*sorted_count].physical_devices)) == VK_INCOMPLETE);
-                }
-
-                if (vkres != VK_SUCCESS) {
-                    loader_instance_heap_free(inst, sorted_array[*sorted_count].physical_devices);
-                    sorted_array[*sorted_count].physical_devices = NULL;
-                    if (vkres == VK_ERROR_OUT_OF_HOST_MEMORY) {
+            // Get the actual physical devices
+            if (0 != count) {
+                do {
+                    sorted_array[*sorted_devices_count].physical_devices =
+                        loader_instance_heap_realloc(inst, sorted_array[*sorted_devices_count].physical_devices,
+                                                     sorted_array[*sorted_devices_count].device_count * sizeof(VkPhysicalDevice),
+                                                     count * sizeof(VkPhysicalDevice), VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);
+                    if (sorted_array[*sorted_devices_count].physical_devices == NULL) {
                         res = VK_ERROR_OUT_OF_HOST_MEMORY;
-                        goto out;
-                    } else {
-                        loader_log(inst, VULKAN_LOADER_WARN_BIT, 0, "Failed to convert DXGI adapter into Vulkan physical device");
-                        continue;
+                        break;
                     }
-                }
-                inst->total_gpu_count += (sorted_array[*sorted_count].device_count = count);
-                sorted_array[*sorted_count].icd_index = icd_idx;
-                sorted_array[*sorted_count].icd_term = icd_term;
-                ++(*sorted_count);
+                    sorted_array[*sorted_devices_count].device_count = count;
+                } while ((vkres = icd_term->scanned_icd->EnumerateAdapterPhysicalDevices(
+                              icd_term->instance, description.AdapterLuid, &count,
+                              sorted_array[*sorted_devices_count].physical_devices)) == VK_INCOMPLETE);
             }
 
-            adapter->lpVtbl->Release(adapter);
+            if (vkres != VK_SUCCESS) {
+                loader_instance_heap_free(inst, sorted_array[*sorted_devices_count].physical_devices);
+                sorted_array[*sorted_devices_count].physical_devices = NULL;
+                if (vkres == VK_ERROR_OUT_OF_HOST_MEMORY) {
+                    res = VK_ERROR_OUT_OF_HOST_MEMORY;
+                    goto out;
+                } else {
+                    loader_log(inst, VULKAN_LOADER_WARN_BIT, 0, "Failed to convert DXGI adapter into Vulkan physical device");
+                    continue;
+                }
+            }
+            sorted_array[*sorted_devices_count].device_count = count;
+            sorted_array[*sorted_devices_count].icd_index = icd_idx;
+            sorted_array[*sorted_devices_count].icd_term = icd_term;
+            (*sorted_devices_count)++;
         }
 
-        dxgi_factory->lpVtbl->Release(dxgi_factory);
+        adapter->lpVtbl->Release(adapter);
     }
 
-out:
+    dxgi_factory->lpVtbl->Release(dxgi_factory);
 
-    if (*sorted_count == 0 && *sorted_devices != NULL) {
+out:
+    if (*sorted_devices_count == 0 && *sorted_devices != NULL) {
         loader_instance_heap_free(inst, *sorted_devices);
         *sorted_devices = NULL;
     }
+    *sorted_devices_count = *sorted_devices_count;
+    *sorted_devices = *sorted_devices;
     return res;
 }
 
@@ -912,4 +929,74 @@ VkLoaderFeatureFlags windows_initialize_dxgi(void) {
     return feature_flags;
 }
 
+// Sort the VkPhysicalDevices that are part of the current group with the list passed in from the sorted list.
+// Multiple groups could have devices out of the same sorted list, however, a single group's devices must all come
+// from the same sorted list.
+void windows_sort_devices_in_group(struct loader_instance *inst, struct VkPhysicalDeviceGroupProperties *group_props,
+                                   struct loader_phys_dev_per_icd *icd_sorted_list) {
+    uint32_t cur_index = 0;
+    for (uint32_t dev = 0; dev < icd_sorted_list->device_count; ++dev) {
+        for (uint32_t grp_dev = cur_index; grp_dev < group_props->physicalDeviceCount; ++grp_dev) {
+            if (icd_sorted_list->physical_devices[dev] == group_props->physicalDevices[grp_dev]) {
+                if (cur_index != grp_dev) {
+                    VkPhysicalDevice swap_dev = group_props->physicalDevices[cur_index];
+                    group_props->physicalDevices[cur_index] = group_props->physicalDevices[grp_dev];
+                    group_props->physicalDevices[grp_dev] = swap_dev;
+                }
+                cur_index++;
+                break;
+            }
+        }
+    }
+    if (cur_index == 0) {
+        loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
+                   "windows_sort_devices_in_group:  Never encountered a device in the sorted list group");
+    }
+}
+
+// This function sorts an array in physical device groups based on the sorted physical device information
+VkResult windows_sort_physical_device_groups(struct loader_instance *inst, const uint32_t group_count,
+                                             struct loader_physical_device_group_term *sorted_group_term,
+                                             const uint32_t sorted_device_count,
+                                             struct loader_phys_dev_per_icd *sorted_phys_dev_array) {
+    if (0 == group_count || NULL == sorted_group_term) {
+        loader_log(inst, VULKAN_LOADER_WARN_BIT, 0,
+                   "windows_sort_physical_device_groups: Called with invalid information (Group count %d, Sorted Info %p)",
+                   group_count, sorted_group_term);
+        return VK_ERROR_INITIALIZATION_FAILED;
+    }
+
+    uint32_t new_index = 0;
+    for (uint32_t icd = 0; icd < sorted_device_count; ++icd) {
+        for (uint32_t dev = 0; dev < sorted_phys_dev_array[icd].device_count; ++dev) {
+            // Find a group associated with a given device
+            for (uint32_t group = new_index; group < group_count; ++group) {
+                bool device_found = false;
+                // Look for the current sorted device in a group and put it in the correct location if it isn't already
+                for (uint32_t grp_dev = 0; grp_dev < sorted_group_term[group].group_props.physicalDeviceCount; ++grp_dev) {
+                    if (sorted_group_term[group].group_props.physicalDevices[grp_dev] ==
+                        sorted_phys_dev_array[icd].physical_devices[dev]) {
+                        // First, sort devices inside of group to be in priority order
+                        windows_sort_devices_in_group(inst, &sorted_group_term[group].group_props, &sorted_phys_dev_array[icd]);
+
+                        // Second, move the group up in priority if it needs to be
+                        if (new_index != group) {
+                            struct loader_physical_device_group_term tmp = sorted_group_term[new_index];
+                            sorted_group_term[new_index] = sorted_group_term[group];
+                            sorted_group_term[group] = tmp;
+                        }
+                        device_found = true;
+                        new_index++;
+                        break;
+                    }
+                }
+                if (device_found) {
+                    break;
+                }
+            }
+        }
+    }
+    return VK_SUCCESS;
+}
+
 #endif  // _WIN32
\ No newline at end of file
